
/*
 * Copyright (c) 2006-2009 by Roland Riegel <feedback@roland-riegel.de>
 *
 * This file is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */
#include <stdlib.h>
#include <stdio.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <ctype.h>
#include <string.h>

//#include <avr/sleep.h>

#include "fat.h"
#include "fat_config.h"
#include "partition.h"
#include "sd_raw.h"
#include "sd_raw_config.h"
#include "main.h"
#include "ADXL345.h"

#define DEBUG 0
#define EEaddress 0x0100//EEPROM address for ADC compare values

//Camera Functions
void send_cmd(uint8_t cmd[]);
void flush(void);
uint8_t check_recv(uint8_t check[], uint8_t offset);

static uint8_t find_file_in_dir(struct fat_fs_struct* fs, struct fat_dir_struct* dd, const char* name, struct fat_dir_entry_struct* dir_entry);
static struct fat_file_struct* open_file_in_dir(struct fat_fs_struct* fs, struct fat_dir_struct* dd, const char* name); 
uint8_t print_disk_info(const struct fat_fs_struct* fs);

char init_filesystem(void);
struct fat_file_struct* create_log_file(void);
void save_data(struct fat_file_struct **file_handle, const char * const buf);

void ioinit(void);

struct fat_dir_struct* dd;
struct fat_dir_entry_struct directory;
struct fat_fs_struct* fs;
struct partition_struct* partition;

struct fat_file_struct * log_file_handle;
char low_byte=0;
char high_byte=0;
int x_value=0, y_value=0, z_value=0;
char ascii_value[6];
char data_buffer[512];

//===========//
//Global Vars//
//===========//
uint8_t sync[6] = {0xAA, 0x0D, 0x00, 0x00, 0x00, 0x00};
uint8_t ack[6] = {0xAA, 0x0E, 0x0D, 0x00, 0x00, 0x00};

volatile uint8_t in[256];
volatile uint8_t in_position = 0;
volatile uint8_t status = 0;


ISR(USART_RX_vect)
{
	cli();
	
	
	in[in_position++] = UDR0;
	if(in_position==12)status = 1;
	
	sei();
}

int main(void)
{
	uint8_t f1=0;
	uint8_t f2=0;

    /* setup uart */
    //uart_init();

	ioinit();	//Setup the AVR I/O for data logging
	flush();
	
	//Initialize SD and FAT32
	while(!init_filesystem());
        
	//Flash lights to indicate succesfull initialization
	LED1_ON();
	delay_ms(100);		
	LED1_OFF();
	
	LED2_ON();
	delay_ms(100);		
	LED2_OFF();
	
	LED3_ON();
	delay_ms(100);		
	LED3_OFF();	
	delay_ms(100);		
		
	//Print some card information as a boot message
	print_disk_info(fs);

	//Initialize system for file logging
	log_file_handle=create_log_file();
	
	sei();
	while(1){
	
		while(status!=1)
		{
			send_cmd(sync);
			delay_ms(20);
			in_position=0;	//Reset the buffer so new characters land at the beginning
		}	
		cli();		//If status is 1, we've received 12 characters from the camera. Shut off interrupts to prevent the buffer from being accidentally changed
		status=0;	//Reset the status flag
		
		f1 = check_recv(ack,0);
		f2 = check_recv(sync,6);
		
		switch(f1)
		{
			//Successful Ack Received!
			case 0:
				LED1_ON();
				delay_ms(500);
				LED1_OFF();			
				break;
			//Incorrect Ack
			case 1:
				LED2_ON();
				delay_ms(500);
				LED2_OFF();
				break;
		}
		
		switch(f2)
		{
			//Successful Sync Received!
			case 0:
				LED1_ON();
				delay_ms(500);
				LED1_OFF();
				break;
			//Incorrect Sync
			case 1:
				LED2_ON();
				delay_ms(500);
				LED2_OFF();			
				break;
		}		
		//if(strlen(data_buffer)>=495){
		//	LED1_ON();
		//	save_data(&log_file_handle, data_buffer);
		//	*data_buffer='\0';
		//	LED1_OFF();
		//}
	
	}
	
	/* provide a simple shell */
	//char buffer[24];
	//while(1)
	//{
	
		/* read command */
		//char* command = buffer;
		//if(read_line(command, sizeof(buffer)) < 1)
		//    continue;

		/* execute command */
	//}

	/* close file system */
	//fat_close(fs);

	/* close partition */
	//partition_close(partition);

    return 0;
}

uint8_t find_file_in_dir(struct fat_fs_struct* fs, struct fat_dir_struct* dd, const char* name, struct fat_dir_entry_struct* dir_entry)
{
    while(fat_read_dir(dd, dir_entry))
    {
        if(strcmp(dir_entry->long_name, name) == 0)
        {
            fat_reset_dir(dd);
            return 1;
        }
    }

    return 0;
}

struct fat_file_struct* open_file_in_dir(struct fat_fs_struct* fs, struct fat_dir_struct* dd, const char* name)
{
    struct fat_dir_entry_struct file_entry;
    if(!find_file_in_dir(fs, dd, name, &file_entry))
        return 0;

    return fat_open_file(fs, &file_entry);
}

//Function: ioinit
//Purpose:	Initialize AVR I/O, UART and Interrupts
//Inputs:	None
//Outputs:	None
void ioinit(void)
{
    //1 = output, 0 = input
	
	adxl345_hw_setup();
	
	DDRD |= (1<<TX_O)|(1<<STAT3)|(1<<STAT2)|(1<<STAT1);		//Define Port D Outputs
	PORTD |= (1<<RX_I);		//Use internal pull-ups on these inputs
    DDRD &= ~(1<<RX_I);	//Define Port D inputs

	// USART Baud rate: 57600 (With 8 MHz Clock)
	
    UBRR0H = (MYUBRR >> 8) & 0x7F;	//Make sure highest bit(URSEL) is 0 indicating we are writing to UBRRH
	UBRR0L = MYUBRR;
    UCSR0A = (1<<U2X0);					//Double the UART Speed
	UCSR0B = (1<<RXEN0)|(1<<TXEN0)|(1<<RXCIE0);
    UCSR0C = (1<<UCSZ00)|(1<<UCSZ01);		//8-Bit Characters
    stdout = &mystdout; //Required for printf init
	
	// Init timer 2
	//Set Prescaler to 8. (Timer Frequency set to 1Mhz
	TCCR2B = (1<<CS21); 	//Divde clock by 8 for 1 Mhz Timer 2 Frequency
    
	//Initialize the External Interrupts
	
	
	cli();
}

static int uart_putchar(char c, FILE *stream)
{
  if (c == '\n')
    uart_putchar('\r', stream);
  
  loop_until_bit_is_set(UCSR0A, UDRE0);
  UDR0 = c;
  return 0;
}

uint8_t uart_getchar(void)
{
    while( !(UCSR0A & (1<<RXC0)) );
	return(UDR0);
}

struct fat_file_struct* create_log_file(void)
{
	struct fat_dir_entry_struct file_entry;
	char name[32];
	int file_number=0;
	
	sprintf(name, "ADXL345_%03d.csv", file_number);
	while(find_file_in_dir(fs,dd,name,&directory)){
		file_number++;
		sprintf(name, "ADXL345_%03d.csv", file_number);
	}
	
#ifdef DEBUG
	if(!fat_create_file(dd, name, &file_entry))
	{
		printf_P(PSTR("No File Created!\n\r"));
	}
#endif
   /* search file in current directory and open it */
	return open_file_in_dir(fs, dd, name);

}

//Usage: saveData(FILE_NAME, log_data, log_data_index);
//Inputs: char *buf - character array to be saved
//		  int buf_size - size of character array
//Output: buffer array is saved to LOG_FILE
//Description: Saves the buf character array to the SD card.
//CONDITIONS: LOG_FILE must be initialized to the handle of an open file.
void save_data(struct fat_file_struct **file_handle, const char * const buf)
{
	fat_write_file(*file_handle, (unsigned char*)buf, strlen(buf));
	sd_raw_sync();
}

char init_filesystem(void)
{
	//setup sd card slot 
	if(!sd_raw_init())
	{
#if DEBUG
		printf_P(PSTR("MMC/SD initialization failed\n"));
#endif
		return 0;
	}

	//open first partition
	partition = partition_open(sd_raw_read,
									sd_raw_read_interval,
#if SD_RAW_WRITE_SUPPORT
									sd_raw_write,
									sd_raw_write_interval,
#else
									0,
									0,
#endif
									0
							   );

	if(!partition)
	{
		//If the partition did not open, assume the storage device
		//is a "superfloppy", i.e. has no MBR.
		partition = partition_open(sd_raw_read,
								   sd_raw_read_interval,
#if SD_RAW_WRITE_SUPPORT
								   sd_raw_write,
								   sd_raw_write_interval,
#else
								   0,
								   0,
#endif
								   -1
								  );
		if(!partition)
		{
#if DEBUG
			printf_P(PSTR("opening partition failed\n"));
#endif
			return 0;
		}
	}

	//Open file system
	//struct fat_fs_struct* fs = fat_open(partition);
	fs = fat_open(partition);
	if(!fs)
	{
#if DEBUG
		printf_P(PSTR("opening filesystem failed\n"));
#endif
		return 0;
	}

	//Open root directory
	//struct fat_dir_entry_struct directory;
	fat_get_dir_entry_of_path(fs, "/", &directory);

	//struct fat_dir_struct* dd = fat_open_dir(fs, &directory);
	dd=fat_open_dir(fs, &directory);
	if(!dd)
	{
#if DEBUG
		printf_P(PSTR("opening root directory failed\n"));
#endif
		return 0;
	}
	return 1;
}

//General short delays
void delay_ms(uint16_t x)
{
    for (; x > 0 ; x--)
        delay_us(1000);
}

//General short delays
void delay_us(uint16_t x)
{    
    while(x > 256)
    {
        TIFR2 = (1<<TOV2); //Clear any interrupt flags on Timer2
        TCNT2 = 0; //256 - 125 = 131 : Preload timer 2 for x clicks. Should be 1us per click
        while( (TIFR2 & (1<<TOV2)) == 0);
       
        x -= 256;
    }

    TIFR2 = (1<<TOV2); //Clear any interrupt flags on Timer2
    TCNT2= 256 - x; //256 - 125 = 131 : Preload timer 2 for x clicks. Should be 1us per click
    while( (TIFR2 & (1<<TOV2)) == 0);
} 
void flush(void)
{
	for(int i=0; i<256; i++)
	{
		in[i] = 0xff;
	}
}

void send_cmd(uint8_t cmd[])
{
	for(int i=0; i<6; i++)
	{
		while((UCSR0A&(1<<UDRE0))==0);
		//loop_until_bit_is_set(UCSR0A, UDRE0);
		UDR0 = cmd[i];
	}
	
}
uint8_t check_recv(uint8_t check[], uint8_t offset)
{
	uint8_t temp[6];
	uint8_t x=0;
	
	for(int i=offset; i<offset+6; i++)
	{
		temp[x] = in[i];
		x++;
	}
	
	if(temp[0]!=0xAA) return 1;
	
	if(temp[1]!=check[1]) return 1;
	
	if(temp[2]!=check[2]) return 1;
	
	return 0;
}

uint8_t print_disk_info(const struct fat_fs_struct* fs)
{
    if(!fs){
		printf_P(PSTR("No File System\n\r"));
        return 0;
	}

    struct sd_raw_info disk_info;
    if(!sd_raw_get_info(&disk_info)){
		printf_P(PSTR("No Info Avaliable\n\r"));
        return 0;
	}

	
    printf("manuf:  0x%x\n\r", disk_info.manufacturer);
    /*
	uart_puts_p(PSTR("oem:    ")); uart_puts((char*) disk_info.oem); uart_putc('\n');
    uart_puts_p(PSTR("prod:   ")); uart_puts((char*) disk_info.product); uart_putc('\n');
    uart_puts_p(PSTR("rev:    ")); uart_putc_hex(disk_info.revision); uart_putc('\n');
    uart_puts_p(PSTR("serial: 0x")); uart_putdw_hex(disk_info.serial); uart_putc('\n');
    uart_puts_p(PSTR("date:   ")); uart_putw_dec(disk_info.manufacturing_month); uart_putc('/');
                                   uart_putw_dec(disk_info.manufacturing_year); uart_putc('\n');
    uart_puts_p(PSTR("size:   ")); uart_putdw_dec(disk_info.capacity / 1024 / 1024); uart_puts_p(PSTR("MB\n"));
    uart_puts_p(PSTR("copy:   ")); uart_putw_dec(disk_info.flag_copy); uart_putc('\n');
    uart_puts_p(PSTR("wr.pr.: ")); uart_putw_dec(disk_info.flag_write_protect_temp); uart_putc('/');
                                   uart_putw_dec(disk_info.flag_write_protect); uart_putc('\n');
    uart_puts_p(PSTR("format: ")); uart_putw_dec(disk_info.format); uart_putc('\n');
	
    printf_P(PSTR("free:   \n\r"));
	printf("%x/%x\n", fat_get_fs_free(fs),fat_get_fs_size(fs));
    */
	return 1;
}