
/*
 JPEG trigger board code
 Sparkfun Electronics
 Viliam Klein
 
 FAT filesystem credit:
 
 Copyright (c) 2006-2009 by Roland Riegel <feedback@roland-riegel.de>
 
 This file is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License version 2 as
 published by the Free Software Foundation. 
*/
 
 
#include <stdlib.h>
#include <stdio.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <ctype.h>
#include <string.h>

#include "fat.h"
#include "fat_config.h"
#include "partition.h"
#include "sd_raw.h"
#include "sd_raw_config.h"
#include "main.h"

#define DEBUG 1


//=Camera Functions=//
void send_cmd(char cmd[]);
uint8_t check_recv(char check[], uint8_t offset);
void sync_camera(void);
void take_pic(void);

void flush(void);

void ioinit(void);
void blink(void);

//=FAT file system functions=//
static uint8_t find_file_in_dir(struct fat_fs_struct* fs, struct fat_dir_struct* dd, const char* name, struct fat_dir_entry_struct* dir_entry);
static struct fat_file_struct* open_file_in_dir(struct fat_fs_struct* fs, struct fat_dir_struct* dd, const char* name);
char init_filesystem(void);
struct fat_file_struct* create_log_file(void);
struct fat_file_struct* create_pic_file(int id, int filenumber);
void save_data(struct fat_file_struct **file_handle, const char * const buf, uint16_t data_length);

//===FAT structs===//
struct fat_dir_struct* dd;
struct fat_dir_entry_struct directory;
struct fat_fs_struct* fs;
struct partition_struct* partition;
struct fat_file_struct* pic_file_handle;

//===Global Vars===//
char sync[6] = {0xAA, 0x0D, 0x00, 0x00, 0x00, 0x00};
char ack[6] = {0xAA, 0x0E, 0x0D, 0x00, 0x00, 0x00};

volatile char in[256];//array used for serial RX
char datab[128];//array used for picture package received from camera

volatile uint8_t in_position = 0;//position in in array
volatile uint8_t status = 0;//flag used to signal that camera is communicating
volatile int file0 = 0;//file name counter for int0
volatile int file1 = 0;//file name counter for int1

//Handler used to store received serial data
ISR(USART_RX_vect)
{
	cli();	
	
	in[in_position++] = UDR0;
	if(in_position==12)status = 1;
	
	sei();
}

//Handler used for int0 vector
SIGNAL(INT0_vect)
{
	cli();
	
	char ini[6] = {0xAA,0x01,0x00,0x07,0x00,0x07};//Use this array to adjust the type of picture; however you will have to change other parts of the code for other picture modes to work
	char pac[6] = {0xAA,0x06,0x08,0x86,0x00,0x00};	
	char comp[6] = {0xAA,0x05,0x00,0x00,0x00,0x00};
	
	LEDB_OFF();
	
	//Disable external interrupts to prevent another trigger while on is being handled.
	EICRA = 0b00000000;
	EIMSK = 0b00000000;
	
	sei();
	in_position = 0;		
	send_cmd(comp);
	while(in_position < 6);
	cli();
	
	//create picture file
	pic_file_handle = create_pic_file(0, file0);
	file0++;
	 
	//enable intterupts	
	sei();
	
	take_pic();
	
	//reinitialize SD and FAT
	LEDR_ON();	
	while(!init_filesystem());
	LEDR_OFF();
	
	//reready camera for the next picture it will take
	in_position = 0;		
	send_cmd(ini);
	delay_ms(20);
	
	//set package size to 128 bytes
	in_position = 0;
	send_cmd(pac);
	while(in_position < 6);
	delay_ms(1);
	
	
	//enable external interrupts
	PORTD |= 0b00001100;
	EICRA = 0b00000000;	// Low level generates interrupt
	EIMSK = 0b00000011;	// External Interrupt Request Disable
	
	sei();
}

SIGNAL(INT1_vect)
{
	cli();
	
	char ini[6] = {0xAA,0x01,0x00,0x07,0x00,0x07};//Use this array to adjust the type of picture; however you will have to change other parts of the code for other picture modes to work
	char pac[6] = {0xAA,0x06,0x08,0x86,0x00,0x00};	
	char comp[6] = {0xAA,0x05,0x00,0x00,0x00,0x00};
	
	LEDB_OFF();
	
	//Disable external interrupts to prevent another trigger while on is being handled.
	EICRA = 0b00000000;		
	EIMSK = 0b00000000;		
	
	sei();
	in_position = 0;		
	send_cmd(comp);
	while(in_position < 6);
	cli();
	
	//create picture file
	pic_file_handle = create_pic_file(1, file1);
	file1++;
	
	//enable intterupts	
	sei();
	
	take_pic();
	
	//reinitialize SD and FAT
	LEDR_ON();	
	while(!init_filesystem());
	LEDR_OFF();
	
	in_position = 0;		
	send_cmd(ini);
	delay_ms(20);
	
	//set package size to 128 bytes
	in_position = 0;
	send_cmd(pac);
	while(in_position < 6);
	delay_ms(1);
	
	
	//enable external interrupts
	PORTD |= 0b00001100;
	EICRA = 0b00000000;	// Low level generates interrupt
	EIMSK = 0b00000011;	// External Interrupt Request Disable
	
	sei();
}

int main(void)
{
	char ini[6] = {0xAA,0x01,0x00,0x07,0x00,0x07};//Initialize JPEG Snapshot Picture for 640x480 resolution. //Use this array to adjust the type of picture; however you will have to change other parts of the code for other picture modes to work
	char pac[6] = {0xAA,0x06,0x08,0x86,0x00,0x00};//set package size to 128 bytes
	
	ioinit();//enable io pins
	flush();//flush in array
	
	LEDR_ON();
	LEDG_ON();
	LEDB_ON();
	
	delay_ms(1000);
	
	LEDR_OFF();
	LEDG_OFF();
	LEDB_OFF();

	//Initialize SD and FAT
	LEDR_ON();	
	while(!init_filesystem());
	LEDR_OFF();
		
	//indicate SD card is good.
	LEDG_ON();
	LEDB_OFF();
   
	//enable intterupts
	sei();
	
	//send sync command initalize camera unitl camera responds.
	//If camera does not respond in 60 syncs, red LED will blink forever
	//also disables interrupts	
	sync_camera();
	LEDG_ON();
	
	sei();
	
	//send Initilize command to camera
	//JPEG snapshot, 640x480
	in_position = 0;		
	send_cmd(ini);
	delay_ms(20);
	
	//set package size to 128 bytes
	in_position = 0;
	send_cmd(pac);
	while(in_position < 6);
	delay_ms(1);
	
	LEDG_OFF();
	
	// Set up external interrupts
	PORTD |= 0b00001100;
	EICRA = 0b00000000;	// Low level generates interrupt
	EIMSK = 0b00000011;	// External Interrupt Request Enable
	
	//loop forever waiting for triggers
	//the program needs to send an ack to the camera to prevent it from sleeping while the avr waits for a trigger
	while(1)
	{
		LEDB_OFF();
		send_cmd(ack);
		delay_ms(100);
		send_cmd(ack);
		delay_ms(100);
		send_cmd(ack);
		delay_ms(100);
		send_cmd(ack);
		delay_ms(100);
		LEDB_ON();
		send_cmd(ack);
		delay_ms(100);
		send_cmd(ack);
		delay_ms(100);
		send_cmd(ack);
		delay_ms(100);
		send_cmd(ack);
		delay_ms(100);
	}	
	
	return 0;
}

//This function handles the data received from the camera
void take_pic(void)
{
	uint8_t byte1, byte2;
	uint16_t datasize, picsize, framecount;	
	char pic[6] = {0xAA,0x04,0x01,0x00,0x00,0x00};	
	
	//Get snapshot picture
	in_position = 0;
	sei();		
	send_cmd(pic);
	while(in_position < 12);
	delay_ms(10);
	LEDR_OFF();
	
	
	cli();//disable interrupts to prevent receive buffer from changing
	//calculate the number of frames (data packets)
	byte1=in[10];
	byte2=in[9];
	picsize = (byte1<<8);
	picsize = (picsize|byte2);
	framecount = picsize/128;
	
	//Begin to capture data	
	in_position = 0;
	ack[2] = 0;	
	ack[3] = 0;
	ack[4] = 0;
	sei();
	send_cmd(ack);//tells camera we are ready to receive first data packet	
	//while(in_position < 120); //wait for camera to send  DATA(6 bytes) + first packet (134 bytes)
	delay_ms(30);
	
	//calculated datasize of packet recieved (should be 128)
	byte1=in[3];
	byte2=in[2];
	datasize = (byte1<<8);
	datasize = (datasize|byte2);
	
	//move picture data out of receive buffer and into the data buffer that will be passed to the SD card
	for(int j=0; j<datasize; j++)
	{		
		datab[j] = in[j+4];		
	}
		
	//save data to SD card file here.
	save_data(&pic_file_handle, datab, datasize);
		
	in_position = 0;//reset receive buffer position
	ack[4] = 1;
	ack[5] = 0;
	sei(); //reenable interruts
	send_cmd(ack); //get the second picture frame	
	while(in_position < 133);
	delay_ms(10);
	
	//loop to receive remaining JPEG packets
	for(int i=2; i<framecount+1; i++)
	{			
		//calculated datasize of packet recieved (should be 128)
		byte1=in[3];
		byte2=in[2];
		datasize = (byte1<<8);
		datasize = (datasize|byte2);
		
		//move picture data out of receive buffer and into the data buffer that will be passed to the SD card
		for(int j=0; j<datasize; j++)
		{			
			datab[j] = in[j+4];			
		}
		
		//save data buffer to SD file
		save_data(&pic_file_handle, datab, datasize);		
		
		//calculate next data packet number
		byte1 = (i&0x00FF);
		byte2 = ((i&0xFF00)>>8); 		
		ack[4] = byte1;
		ack[5] = byte2;
		
		in_position = 0;	
		send_cmd(ack);					
		//while(in_position < 100);
		delay_ms(30);		
	}
	
	delay_ms(20);
	
	//get final picture packet
	ack[4] = 0xf0;
	ack[5] = 0xf0;
	
	in_position = 0;	
	send_cmd(ack);					
	//while(in_position < 100);
	delay_ms(20);
	
	//calculated datasize of packet recieved (should be 128)
	byte1=in[3];
	byte2=in[2];
	datasize = (byte1<<8);
	datasize = (datasize|byte2);
	
	//move picture data out of receive buffer and into the data buffer that will be passed to the SD card
	for(int j=0; j<datasize; j++)
	{			
		datab[j] = in[j+4];			
	}
	
	//save data buffer to SD file
	save_data(&pic_file_handle, datab, datasize);
	
	
	cli();//disabel interrupts
	
	//close file and filesystem
	fat_close_file(pic_file_handle);//close picture file
	fat_close_dir(dd);
	fat_close(fs);//close file system	
	partition_close(partition);//close partition
	unselect_card();
}


//sends sync command to talk to camera
void sync_camera(void)
{
	uint8_t x=0;
	
	
	//send sync command until connection established or 60 syncs sent (see camera datasheet)
	while(status!=1 && x<60)
	{
		send_cmd(sync);
		delay_ms(50);
		x++;
	}
	
	//disable interrupts to allow checking of receive buffer without interruption
	cli();		
	status=0;
	
	//if receive buffer does not contain correct ACK, blink forever
	if(check_recv(ack,0))
	{
		LEDG_OFF();
		LEDB_OFF();
		
		while(1)
		{
			blink();
		}
	}
}

void blink(void)
{

	for(int x =0; x<40; x++){
	LEDR_ON(); delay_ms(200);
	LEDR_OFF(); delay_ms(200);
	}

}

uint8_t find_file_in_dir(struct fat_fs_struct* fs, struct fat_dir_struct* dd, const char* name, struct fat_dir_entry_struct* dir_entry)
{
    while(fat_read_dir(dd, dir_entry))
    {
        if(strcmp(dir_entry->long_name, name) == 0)
        {
            fat_reset_dir(dd);
            return 1;
        }
    }

    return 0;
}

struct fat_file_struct* open_file_in_dir(struct fat_fs_struct* fs, struct fat_dir_struct* dd, const char* name)
{
    struct fat_dir_entry_struct file_entry;
    if(!find_file_in_dir(fs, dd, name, &file_entry))
        return 0;

    return fat_open_file(fs, &file_entry);
}

//Function: ioinit
//Purpose:	Initialize AVR I/O, UART and Interrupts
void ioinit(void)
{
    //1 = output, 0 = input
	//1 = output, 0 = input
    DDRB = 0b11101110; //MISO on PB4, CD on PB0
    DDRC = 0b10111111; //
    DDRD = 0b11110010; //RX on PD0, INT0 and INT1 on 2 and 3
	
	DDRD |= (1<<TX_O)|(1<<STAT3)|(1<<STAT2)|(1<<STAT1);//Define Port D Outputs
	PORTD |= (1<<RX_I);//Use internal pull-ups on these inputs
    DDRD &= ~(1<<RX_I);//Define Port D inputs

	// USART Baud rate: 57600 (With 8 MHz Clock)
	
    UBRR0H = (MYUBRR >> 8) & 0x7F;	//Make sure highest bit(URSEL) is 0 indicating we are writing to UBRRH
	UBRR0L = MYUBRR;
    UCSR0A = (1<<U2X0);//Double the UART Speed
	UCSR0B = (1<<RXEN0)|(1<<TXEN0)|(1<<RXCIE0);
    UCSR0C = (1<<UCSZ00)|(1<<UCSZ01);//8-Bit Characters
    stdout = &mystdout; //Required for printf init
	
	TCCR2B = (1<<CS21);
}

static int uart_putchar(char c, FILE *stream)
{
  if (c == '\n')
    uart_putchar('\r', stream);
  
  loop_until_bit_is_set(UCSR0A, UDRE0);
  UDR0 = c;
  return 0;
}

uint8_t uart_getchar(void)
{
    while( !(UCSR0A & (1<<RXC0)) );
	return(UDR0);
}

struct fat_file_struct* create_pic_file(int id, int file_number)
{
	struct fat_dir_entry_struct file_entry;
	char name[32];
	
	
	sprintf(name, "%01d_%03d.jpg", id, file_number);
	while(find_file_in_dir(fs,dd,name,&directory)){
		file_number++;
		sprintf(name, "%01d_%03d.jpg", id, file_number);
	}
	
#ifdef DEBUG
	if(!fat_create_file(dd, name, &file_entry))
	{
		printf_P(PSTR("No File Created!\n\r"));
	}
#endif
   /* search file in current directory and open it */
	return open_file_in_dir(fs, dd, name);

}

//Usage: saveData(file_struct
//Inputs: char *buf - character array to be saved
//		  uint16_t data_length - size of character array
//Output: buffer array is saved to picture file
//Description: Saves data array to the SD card.
//CONDITIONS: picture file must be initialized to the handle of an open file.
void save_data(struct fat_file_struct **file_handle, const char * const buf,uint16_t data_length)
{
	LEDG_ON();
	fat_write_file(*file_handle, (unsigned char*)buf, data_length);
	sd_raw_sync();
	LEDG_OFF();
}

char init_filesystem(void)
{
	//setup sd card slot 
	if(!sd_raw_init())
	{
#if DEBUG
	//	printf_P(PSTR("MMC/SD initialization failed\n"));
#endif
		return 0;
	}

	//open first partition
	partition = partition_open(sd_raw_read,
									sd_raw_read_interval,
#if SD_RAW_WRITE_SUPPORT
									sd_raw_write,
									sd_raw_write_interval,
#else
									0,
									0,
#endif
									0
							   );

	//printf("p1\n");
	
	if(!partition)
	{
		//If the partition did not open, assume the storage device
		//is a "superfloppy", i.e. has no MBR.
		partition = partition_open(sd_raw_read,
								   sd_raw_read_interval,
#if SD_RAW_WRITE_SUPPORT
								   sd_raw_write,
								   sd_raw_write_interval,
#else
								   0,
								   0,
#endif
								   -1
								  );
		if(!partition)
		{
#if DEBUG
		//	printf_P(PSTR("opening partition failed\n"));
#endif
			return 0;
		}
	}
	
	//printf("p2\n");
	
	//Open file system
	//struct fat_fs_struct* fs = fat_open(partition);
	fs = fat_open(partition);
	if(!fs)
	{
#if DEBUG
	//	printf_P(PSTR("opening filesystem failed\n"));
#endif
		return 0;
	}
	
	//printf("p3\n");
	
	//Open root directory
	//struct fat_dir_entry_struct directory;
	fat_get_dir_entry_of_path(fs, "/", &directory);

	//struct fat_dir_struct* dd = fat_open_dir(fs, &directory);
	dd=fat_open_dir(fs, &directory);
	if(!dd)
	{
#if DEBUG
	//	printf_P(PSTR("opening root directory failed\n"));
#endif
		return 0;
	}
	
	//printf("p4\n");
	
	return 1;
}

//General short delays
void delay_ms(uint16_t x)
{
    for (; x > 0 ; x--)
        delay_us(1000);
}

//General short delays
void delay_us(uint16_t x)
{    
    while(x > 256)
    {
        TIFR2 = (1<<TOV2); //Clear any interrupt flags on Timer2
        TCNT2 = 0; //256 - 125 = 131 : Preload timer 2 for x clicks. Should be 1us per click
        while( (TIFR2 & (1<<TOV2)) == 0);
       
        x -= 256;
    }

    TIFR2 = (1<<TOV2); //Clear any interrupt flags on Timer2
    TCNT2= 256 - x; //256 - 125 = 131 : Preload timer 2 for x clicks. Should be 1us per click
    while( (TIFR2 & (1<<TOV2)) == 0);
} 
void flush(void)
{
	for(int i=0; i<134; i++)
	{
		in[i] = 0xff;
	}
	for(int i=0; i<128; i++)
	{
		datab[i] = 0xff;
	}
}

//function used to send commands to camera
void send_cmd(char cmd[])
{
	cli();
	
	for(int i=0; i<6; i++)
	{
		while((UCSR0A&(1<<UDRE0))==0);
		//loop_until_bit_is_set(UCSR0A, UDRE0);
		UDR0 = cmd[i];
	}
	
	sei();
	
}

//function that checks to make sure correct bytes are received from camera
uint8_t check_recv(char check[], uint8_t offset)
{
	uint8_t temp[6];
	uint8_t x=0;
	
	for(int i=offset; i<offset+6; i++)
	{
		temp[x] = in[i];
		x++;
	}
	
	if(temp[0]!=0xAA) return 1;
	
	if(temp[1]!=check[1]) return 1;
	
	if(temp[2]!=check[2]) return 1;
	
	return 0;
}
